@@ -19,8 +19,8 @@ local sysdetect = require("luarocks.core.sysdetect")
 
 --------------------------------------------------------------------------------
 
-local program_version = "3.2.1"
-local program_series = "3.2"
+local program_version = "3.3.0"
+local program_series = "3.3"
 local major_version = (program_version:match("([^.]%.[^.])")) or program_series
 
 local is_windows = package.config:sub(1,1) == "\\"
@@ -187,11 +187,11 @@ local function make_defaults(lua_version, target_cpu, platforms, home)
       fs_use_modules = true,
       hooks_enabled = true,
       deps_mode = "one",
+      no_manifest = false,
       check_certificates = false,
 
       cache_timeout = 60,
       cache_fail_timeout = 86400,
-      version_check_on_fail = true,
 
       lua_modules_path = "/share/lua/"..lua_version,
       lib_modules_path = "/lib/lua/"..lua_version,
@@ -665,8 +665,6 @@ function cfg.init(detected, warning)
    -- Let's finish up the cfg table.
    ----------------------------------------
 
-   cfg.variables.LUA_DIR = detected.given_lua_dir or cfg.variables.LUA_DIR
-
    -- Settings given via the CLI (i.e. --lua-dir) take precedence over config files.
    cfg.project_dir = detected.given_project_dir or cfg.project_dir
    cfg.lua_version = detected.given_lua_version or cfg.lua_version
